#AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"IMG_EXT": "00_utils.ipynb",
         "cd_root": "00_utils.ipynb",
         "extension_counts": "00_utils.ipynb",
         "stop_instance": "00_utils.ipynb",
         "snapshot_protobufs": "01_experiments.ipynb",
         "fetch_experiments_data": "01_experiments.ipynb",
         "nlp": "02_nlp.ipynb",
         "tokenize": "02_nlp.ipynb",
         "lexical_density": "02_nlp.ipynb",
         "Debugger": "03_ml.ipynb",
         "build_pipeline": "03_ml.ipynb",
         "parse_gs_results": "03_ml.ipynb",
         "IMAGENET_STATS": "04_dl.ipynb",
         "gpu_found": "04_dl.ipynb",
         "find_layer_groups": "04_dl.ipynb",
         "inverse_sigmoid": "04_dl.ipynb",
         "mish": "04_dl.ipynb",
         "normalize": "04_dl.ipynb",
         "denormalize": "04_dl.ipynb",
         "show_batch": "04_dl.ipynb",
         "show_image_variants": "04_dl.ipynb",
         "ImageSequence": "04_dl.ipynb",
         "CosineLRScheduler": "04_dl.ipynb",
         "BatchHistory": "04_dl.ipynb",
         "PerformanceThreshold": "04_dl.ipynb",
         "plot_history": "04_dl.ipynb",
         "show_predictions": "04_dl.ipynb",
         "hybrid_quantize": "04_dl.ipynb",
         "ClickLiteralOption": "05_cli.ipynb",
         "click.command": "05_cli.ipynb",
         "click.option": "05_cli.ipynb",
         "labeler_agreement": "06_labeling.ipynb",
         "fetch_img_from_url": "07_scraping.ipynb",
         "load_praw_credentials": "07_scraping.ipynb",
         "is_imgur_page": "07_scraping.ipynb",
         "fetch_imgur_url": "07_scraping.ipynb",
         "scrape_subreddit_images_1k": "07_scraping.ipynb",
         "scrape_subreddit_images": "07_scraping.ipynb",
         "remove_corrupt_images": "07_scraping.ipynb"}

modules = ["utils.py",
           "experiments.py",
           "nlp.py",
           "ml.py",
           "dl.py",
           "cli.py",
           "labeling.py",
           "scraping.py"]

git_url = "https://github.com/hdmamin/spellotape/tree/master/"

def custom_doc_links(name): return None